import asyncio
import atexit
import threading
from collections import defaultdict
from contextlib import contextmanager, nullcontext
from typing import Callable, Dict, List, Optional, TextIO

import cursor

from tui_progress.collections import OrderedSet

TERM_MOVE_UP = '\x1b[A'
TERM_CLEAR_LINE = '\x1b[2K'

NOTSET = object()


class RepeatingTimer(threading.Timer):
    """Timer which repeatedly calls a function until cancel() is called"""

    def run(self):
        while not self.finished.wait(self.interval):
            self.function(*self.args, **self.kwargs)


class ProgressIndicator:
    """Represents a text-based progress/status indicator that prints to terminal
    """

    def __init__(self, stream: TextIO, name: str = None, **kwargs):
        super().__init__(**kwargs)
        self.stream: TextIO = stream
        self.name = name
        self.manager: ProgressIndicatorManager = _progress_indicators[self.stream]

        # Whether the indicator needs re-rendering
        self.is_dirty: bool = False
        # Text to render
        self.frame: str = ''
        # Optional frame to print after the indicator is finished
        self.summary_frame: Optional[str] = None

        # Backend used for update_loop
        self.loop_backend: Optional[_UpdateLoopBackend] = None

    def __str__(self):
        name = self.name or '<unnamed>'
        frame = self.summary_frame or self.frame
        return f'{name} {frame!r}'

    def __repr__(self):
        cls = self.__class__.__name__
        return f'<{cls}: {self}>'

    @property
    def is_managed(self):
        return self in self.manager

    def update(self, frame: str):
        """Set the next frame to render, and trigger a refresh"""
        self.frame = frame
        self.is_dirty = True

        if self.is_managed:
            self.manager.trigger_refresh()

    # For compatibility as a file-like object
    write = update

    def start(self):
        """Begin rendering the indicator"""
        self.manager.add(self)

    def stop(self):
        """Stop rendering the indicator"""
        if self.loop_backend:
            self.loop_backend.stop()
            self.loop_backend = None

        self.manager.remove(self)

    def complete(self, frame: str):
        """Stop rendering and print a final frame"""
        self.summary_frame = frame
        self.manager.trigger_refresh()
        self.stop()

    def update_loop(self, get_frame: Callable[[], str], interval: float):
        """Call a method to retrieve the next frame and update every interval"""
        if not self.is_managed:
            # Manager must be started to determine whether the render loop is async
            self.start()

        if self.manager.is_async():
            backend_cls = _AsyncUpdateLoopBackend
        else:
            backend_cls = _SyncUpdateLoopBackend

        self.loop_backend = backend_cls(self, get_frame, interval)
        self.loop_backend.start()


class ProgressIndicatorManager:
    """Handles the rendering of concurrent progress indicators
    """

    def __init__(self, stream: TextIO):
        self._stream = stream

        # Indicators we manage rendering of
        self.indicators: OrderedSet[ProgressIndicator] = OrderedSet()
        self.finished_indicators: OrderedSet[ProgressIndicator] = OrderedSet()
        # Temporary lists used to iterate the managed indicators during a
        # rendering iteration. It is stored as an instance variable to reduce
        # the number of allocations.
        self._rendering_indicators = []
        self._finished_indicators = []

        # Flag used to prevent multiple rendering loops from being started
        self._is_started: bool = False
        # Flag used to determine which rendering loop to use.
        # This value is set to True if a running event loop is found.
        self._async_mode: Optional[bool] = None

        # Handles the details of rendering indicators in a loop (sync vs async)
        self.backend: Optional[_IndicatorRenderLoopBackend] = None

        # Ensure a hidden TTY cursor is re-enabled at exit
        atexit.register(self.show_cursor)

    def __bool__(self):
        return bool(self.indicators)

    def __contains__(self, indicator: ProgressIndicator):
        return indicator in self.indicators

    def is_async(self) -> bool:
        """Return whether the rendering loop is async"""
        return self._async_mode

    def mutation_lock(self):
        if self.backend:
            return self.backend.mutation_lock()
        else:
            return nullcontext()

    def add(self, indicator: ProgressIndicator):
        """Begin management and rendering of an indicator"""
        with self.mutation_lock():
            if indicator not in self.indicators:
                self.indicators.add(indicator)
                self.start()

                # Any additions of indicators should trigger a refresh, as a
                # new line must be rendered.
                self.trigger_refresh()

    def remove(self, indicator: ProgressIndicator):
        """Stop managing and rendering the indicator"""
        with self.mutation_lock():
            if indicator in self.indicators:
                self.indicators.discard(indicator)
                self.finished_indicators.add(indicator)

        ###
        # NOTE: Any removal should trigger a refresh, as all the lines need to be
        #       re-rendered. However, if the last indicator has just been removed,
        #       the Main thread may immediately exit after this call — in that
        #       case, we manually render the final frame, and avoid triggering a
        #       refresh, because it may cause writes to a now-closed stdout/stderr
        #       in another thread
        #
        if not self.indicators:
            self.render_frame()
            self.stop()
        else:
            self.trigger_refresh()

    def trigger_refresh(self):
        """Trigger the rendering of indicators"""
        self.backend.trigger_refresh()

    def start(self):
        if self._is_started:
            return

        try:
            self._async_mode = asyncio.get_running_loop() is not None
        except RuntimeError:
            self._async_mode = False

        if self._async_mode:
            backend_cls = _AsyncIndicatorRenderLoopBackend
        else:
            backend_cls = _SyncIndicatorRenderLoopBackend

        self.backend = backend_cls(self)
        self.backend.start()

        self._is_started = True
        self.hide_cursor()

    def stop(self):
        if not self._is_started:
            return

        self.backend.stop()

        self._is_started = False
        self._async_mode = None
        self.show_cursor()

    def hide_cursor(self):
        if self._stream.isatty():
            cursor.hide(stream=self._stream)

    def show_cursor(self):
        if self._stream.isatty():
            cursor.show(stream=self._stream)

    def render_frame(self):
        indicators: List[ProgressIndicator] = self._rendering_indicators
        finished_indicators: List[ProgressIndicator] = self._finished_indicators

        with self.mutation_lock():
            indicators[:] = self.indicators
            finished_indicators[:] = self.finished_indicators
            self.finished_indicators.clear()

        # print finished summary lines first (don't include in moveup lines)
        # print regular indicators (forced if there were any finished indicators)
        # for every non-summary finished indicator, clear + newline
        # move up len(indicators) + len(non_summary)

        vanished_lines = 0
        for indicator in finished_indicators:
            if indicator.summary_frame is not None:
                self.clear_line()
                self._write(indicator.summary_frame + '\n')
                indicator.summary_frame = None
            else:
                vanished_lines += 1

        forced_render = bool(finished_indicators)
        for indicator in indicators:
            if forced_render or indicator.is_dirty:
                self.clear_line()
                self.render_indicator(indicator)
            else:
                self._write('\n')

        for _ in range(vanished_lines):
            self.clear_line()
            self._write('\n')

        if move_up := len(indicators) + vanished_lines:
            self.moveto(-move_up)

        self._flush()

    def render_indicator(self, indicator: ProgressIndicator, newline: bool = True):
        self._write(indicator.frame + '\n' * newline)
        indicator.is_dirty = False

    def render_summary(self, indicator: ProgressIndicator) -> bool:
        if indicator.summary_frame is not None:
            self._write(indicator.summary_frame)
            indicator.summary_frame = None
            return True
        else:
            return False

    def clear_line(self):
        self._write(TERM_CLEAR_LINE + '\r')

    def moveto(self, n):
        self._write('\n' * n + TERM_MOVE_UP * -n)

    def _write(self, s: str):
        self._stream.write(s)

    def _flush(self):
        self._stream.flush()


class _UpdateLoopBackend:
    def __init__(self, indicator: 'ProgressIndicator', get_frame: Callable[[], str], interval: float):
        self.indicator = indicator
        self.get_frame = get_frame
        self.interval = interval

    def update(self):
        frame = self.get_frame()
        self.indicator.update(frame)

    def start(self):
        raise NotImplementedError

    def stop(self):
        raise NotImplementedError


class _AsyncUpdateLoopBackend(_UpdateLoopBackend):
    def __init__(self, indicator: 'ProgressIndicator', get_frame: Callable[[], str], interval: float):
        super().__init__(indicator, get_frame, interval)
        self.task: Optional[asyncio.Task] = None

    def start(self):
        if self.task is None:
            self.task = asyncio.create_task(self.loop())

    def stop(self):
        if self.task:
            self.task.cancel()
            self.task = None

    async def loop(self):
        try:
            while True:
                self.update()
                await asyncio.sleep(self.interval)
        except asyncio.CancelledError:
            self.update()
            await asyncio.sleep(0)


class _SyncUpdateLoopBackend(_UpdateLoopBackend):
    def __init__(self, indicator: 'ProgressIndicator', get_frame: Callable[[], str], interval: float):
        super().__init__(indicator, get_frame, interval)
        self.timer: Optional[threading.Timer] = None

    def start(self):
        if self.timer is None:
            self.timer = RepeatingTimer(self.interval, self.update)
            self.timer.start()

    def stop(self):
        if self.timer:
            self.timer.cancel()
            self.timer = None


class _IndicatorRenderLoopBackend:
    def __init__(self, manager: 'ProgressIndicatorManager'):
        self.manager = manager

    def render_frame(self):
        self.manager.render_frame()

    @contextmanager
    def mutation_lock(self):
        raise NotImplementedError

    def start(self):
        raise NotImplementedError

    def stop(self):
        raise NotImplementedError

    def trigger_refresh(self):
        raise NotImplementedError


class _AsyncIndicatorRenderLoopBackend(_IndicatorRenderLoopBackend):
    def __init__(self, manager: 'ProgressIndicatorManager'):
        super().__init__(manager)
        self.task: Optional[asyncio.Task] = None
        self.refresh_needed: Optional[asyncio.Event] = None

    @contextmanager
    def mutation_lock(self):
        yield

    def start(self):
        if not self.task:
            self.refresh_needed = asyncio.Event()
            self.task = asyncio.create_task(self.render_loop())

    def stop(self):
        if self.task:
            self.task.cancel()
            self.task = None
            self.refresh_needed = None

    def trigger_refresh(self):
        self.refresh_needed.set()

    async def render_loop(self):
        while True:
            await self.refresh_needed.wait()
            self.render_frame()
            self.refresh_needed.clear()


class _SyncIndicatorRenderLoopBackend(_IndicatorRenderLoopBackend):
    def __init__(self, manager: 'ProgressIndicatorManager'):
        super().__init__(manager)
        self.thread: Optional[threading.Thread] = None

        self.halt: Optional[threading.Event] = None
        self.refresh_needed: Optional[threading.Event] = None
        self.has_event: Optional[threading.Condition] = None

        self.mutex: Optional[threading.Lock] = None

    @contextmanager
    def mutation_lock(self):
        with self.mutex:
            yield

    def start(self):
        if not self.thread:
            self.halt = threading.Event()
            self.refresh_needed = threading.Event()
            self.has_event = threading.Condition()
            self.mutex = threading.Lock()

            self.thread = threading.Thread(target=self.render_loop)
            self.thread.setDaemon(True)
            self.thread.start()

    def stop(self):
        if threading.current_thread() is self.thread:
            # Ensure we don't deadlock acquiring self.has_event by performing
            # stop in another thread.
            request_stop = threading.Thread(target=self.stop)
            request_stop.start()
            request_stop.join()
            return

        if self.thread:
            self.halt.set()
            with self.has_event:
                self.has_event.notify()

            self.thread.join()

            self.thread = None
            self.halt = None
            self.mutex = None

    def trigger_refresh(self):
        self.refresh_needed.set()
        with self.has_event:
            self.has_event.notify()

    def render_loop(self):
        with self.has_event:
            while not self.halt.is_set():
                if self.refresh_needed.is_set():
                    self.render_frame()
                    self.refresh_needed.clear()

                self.has_event.wait()


class ProgressIndicatorStreamManager(defaultdict):
    def __missing__(self, stream: TextIO):
        self[stream] = mgr = ProgressIndicatorManager(stream)
        return mgr


_progress_indicators: Dict[TextIO, ProgressIndicatorManager] = ProgressIndicatorStreamManager()



import colorsys
import math
import statistics
from collections import defaultdict
from contextlib import contextmanager
from typing import Any, Dict, Iterable, Set, Tuple

from cached_property import cached_property
from colorama import Style

try:
    from collections.abc import Sequence  # python >= 3.10
except ImportError:
    from collections import Sequence  # python <= 3.9

try:
    import colormath
    HAS_COLORMATH = True
except ImportError:
    HAS_COLORMATH = False
else:
    from colormath.color_objects import sRGBColor, LabColor
    from colormath.color_conversions import convert_color

CSI = '\033['

NOTSET = object()


def code_to_chars(code):
    return CSI + '38;5;' + str(code) + 'm'


def _require_colormath():
    if not HAS_COLORMATH:
        raise RuntimeError('Installation of the colormath library is required to use this feature.')


class _XtermColorSwatch:
    def __init__(self, color: 'XtermColor'):
        self.color = color

    def __str__(self):
        return self()

    def __repr__(self):
        return str(self)

    def __format__(self, format_spec):
        return format(str(self), format_spec)

    def __call__(self, n: int = 3, c: str = '█'):
        return self.color(c * n)


class XtermColor(str):
    code: int
    rgb: int
    name: str

    def __new__(cls, code: int, rgb: int, name: str):
        chars = code_to_chars(code)
        color = str.__new__(cls, chars)
        color.code = code
        color.rgb = rgb
        color.name = name
        return color

    def __call__(self, s: str = NOTSET) -> str:
        """Return the string surrounded by this color and a style reset
        """
        if s is NOTSET:
            return str(self)
        else:
            return f'{self}{s}{Style.RESET_ALL}'

    def reset(self) -> str:
        """Print the reset-all-styles escape sequence
        """
        return Style.RESET_ALL

    @cached_property
    def swatch(self) -> _XtermColorSwatch:
        return _XtermColorSwatch(self)

    @cached_property
    def r(self) -> int:
        return (self.rgb >> 24) & 0xFF

    @cached_property
    def g(self) -> int:
        return (self.rgb >> 16) & 0xFF

    @cached_property
    def b(self) -> int:
        return self.rgb & 0xFF

    @cached_property
    def red(self) -> float:
        return self.r / 255.0

    @cached_property
    def green(self) -> float:
        return self.g / 255.0

    @cached_property
    def blue(self) -> float:
        return self.b / 255.0

    @cached_property
    def hsv(self) -> Tuple[float, float, float]:
        return colorsys.rgb_to_hsv(self.red, self.green, self.blue)

    @cached_property
    def is_greyscale(self) -> bool:
        hue, saturation, value = self.hsv
        return saturation == 0.0

    @cached_property
    def perceived_brightness(self) -> float:
        """Perceived brightness, according to the HSP colour model

        See http://alienryderflex.com/hsp.html
        """
        return math.sqrt(0.299 * self.red**2 + 0.587 * self.green**2 + 0.114 * self.blue**2)

    @cached_property
    def is_bright(self) -> bool:
        return self.perceived_brightness >= 0.5

    @cached_property
    def is_dark(self) -> bool:
        return self.perceived_brightness < 0.5

    @cached_property
    def as_srgb_color(self) -> 'sRGBColor':
        _require_colormath()
        return sRGBColor(self.red, self.green, self.blue)

    @cached_property
    def as_lab_color(self) -> 'LabColor':
        _require_colormath()
        return convert_color(self.as_srgb_color, LabColor)


class XtermCodes:
    all_colors: Dict[str, XtermColor]
    bright_colors: Dict[str, XtermColor]
    dark_colors: Dict[str, XtermColor]
    by_code: Dict[int, XtermColor]

    def __init__(self):
        self.all_colors = {}
        self.bright_colors = {}
        self.dark_colors = {}
        self.by_code = {}

        for name in dir(self):
            if not name.startswith('_') and name[0].isupper():
                code, rgb = getattr(self, name)
                color = XtermColor(code, rgb, name)
                setattr(self, name, color)

                self.by_code[code] = color
                self.all_colors[name] = color
                if color.is_bright:
                    self.bright_colors[name] = color
                else:
                    self.dark_colors[name] = color

    def __getitem__(self, item):
        if isinstance(item, str):
            return self.all_colors[item]
        else:
            return self.by_code[item]

    def __len__(self):
        return len(self.all_colors)


class XtermFore(XtermCodes):
    differentiated_colors: Tuple[XtermColor]

    BLACK                = 0,   0x000000
    MAROON               = 1,   0x800000
    GREEN                = 2,   0x008000
    OLIVE                = 3,   0x808000
    NAVY                 = 4,   0x000080
    PURPLE               = 5,   0x800080
    TEAL                 = 6,   0x008080
    SILVER               = 7,   0xc0c0c0
    GREY                 = 8,   0x808080
    RED                  = 9,   0xff0000
    LIME                 = 10,  0x00ff00
    YELLOW               = 11,  0xffff00
    BLUE                 = 12,  0x0000ff
    FUCHSIA              = 13,  0xff00ff
    AQUA                 = 14,  0x00ffff
    WHITE                = 15,  0xffffff
    GREY0                = 16,  0x000000
    NAVYBLUE             = 17,  0x00005f
    DARKBLUE             = 18,  0x000087
    BLUE3                = 19,  0x0000af
    BLUE3_1              = 20,  0x0000d7
    BLUE1                = 21,  0x0000ff
    DARKGREEN            = 22,  0x005f00
    DEEPSKYBLUE4         = 23,  0x005f5f
    DEEPSKYBLUE4_1       = 24,  0x005f87
    DEEPSKYBLUE4_2       = 25,  0x005faf
    DODGERBLUE3          = 26,  0x005fd7
    DODGERBLUE2          = 27,  0x005fff
    GREEN4               = 28,  0x008700
    SPRINGGREEN4         = 29,  0x00875f
    TURQUOISE4           = 30,  0x008787
    DEEPSKYBLUE3         = 31,  0x0087af
    DEEPSKYBLUE3_1       = 32,  0x0087d7
    DODGERBLUE1          = 33,  0x0087ff
    GREEN3               = 34,  0x00af00
    SPRINGGREEN3         = 35,  0x00af5f
    DARKCYAN             = 36,  0x00af87
    LIGHTSEAGREEN        = 37,  0x00afaf
    DEEPSKYBLUE2         = 38,  0x00afd7
    DEEPSKYBLUE1         = 39,  0x00afff
    GREEN3_1             = 40,  0x00d700
    SPRINGGREEN3_1       = 41,  0x00d75f
    SPRINGGREEN2         = 42,  0x00d787
    CYAN3                = 43,  0x00d7af
    DARKTURQUOISE        = 44,  0x00d7d7
    TURQUOISE2           = 45,  0x00d7ff
    GREEN1               = 46,  0x00ff00
    SPRINGGREEN2_1       = 47,  0x00ff5f
    SPRINGGREEN1         = 48,  0x00ff87
    MEDIUMSPRINGGREEN    = 49,  0x00ffaf
    CYAN2                = 50,  0x00ffd7
    CYAN1                = 51,  0x00ffff
    DARKRED              = 52,  0x5f0000
    DEEPPINK4            = 53,  0x5f005f
    PURPLE4              = 54,  0x5f0087
    PURPLE4_1            = 55,  0x5f00af
    PURPLE3              = 56,  0x5f00d7
    BLUEVIOLET           = 57,  0x5f00ff
    ORANGE4              = 58,  0x5f5f00
    GREY37               = 59,  0x5f5f5f
    MEDIUMPURPLE4        = 60,  0x5f5f87
    SLATEBLUE3           = 61,  0x5f5faf
    SLATEBLUE3_1         = 62,  0x5f5fd7
    ROYALBLUE1           = 63,  0x5f5fff
    CHARTREUSE4          = 64,  0x5f8700
    DARKSEAGREEN4        = 65,  0x5f875f
    PALETURQUOISE4       = 66,  0x5f8787
    STEELBLUE            = 67,  0x5f87af
    STEELBLUE3           = 68,  0x5f87d7
    CORNFLOWERBLUE       = 69,  0x5f87ff
    CHARTREUSE3          = 70,  0x5faf00
    DARKSEAGREEN4_1      = 71,  0x5faf5f
    CADETBLUE            = 72,  0x5faf87
    CADETBLUE_1          = 73,  0x5fafaf
    SKYBLUE3             = 74,  0x5fafd7
    STEELBLUE1           = 75,  0x5fafff
    CHARTREUSE3_1        = 76,  0x5fd700
    PALEGREEN3           = 77,  0x5fd75f
    SEAGREEN3            = 78,  0x5fd787
    AQUAMARINE3          = 79,  0x5fd7af
    MEDIUMTURQUOISE      = 80,  0x5fd7d7
    STEELBLUE1_1         = 81,  0x5fd7ff
    CHARTREUSE2          = 82,  0x5fff00
    SEAGREEN2            = 83,  0x5fff5f
    SEAGREEN1            = 84,  0x5fff87
    SEAGREEN1_1          = 85,  0x5fffaf
    AQUAMARINE1          = 86,  0x5fffd7
    DARKSLATEGRAY2       = 87,  0x5fffff
    DARKRED_1            = 88,  0x870000
    DEEPPINK4_1          = 89,  0x87005f
    DARKMAGENTA          = 90,  0x870087
    DARKMAGENTA_1        = 91,  0x8700af
    DARKVIOLET           = 92,  0x8700d7
    PURPLE_1             = 93,  0x8700ff
    ORANGE4_1            = 94,  0x875f00
    LIGHTPINK4           = 95,  0x875f5f
    PLUM4                = 96,  0x875f87
    MEDIUMPURPLE3        = 97,  0x875faf
    MEDIUMPURPLE3_1      = 98,  0x875fd7
    SLATEBLUE1           = 99,  0x875fff
    YELLOW4              = 100, 0x878700
    WHEAT4               = 101, 0x87875f
    GREY53               = 102, 0x878787
    LIGHTSLATEGREY       = 103, 0x8787af
    MEDIUMPURPLE         = 104, 0x8787d7
    LIGHTSLATEBLUE       = 105, 0x8787ff
    YELLOW4_1            = 106, 0x87af00
    DARKOLIVEGREEN3      = 107, 0x87af5f
    DARKSEAGREEN         = 108, 0x87af87
    LIGHTSKYBLUE3        = 109, 0x87afaf
    LIGHTSKYBLUE3_1      = 110, 0x87afd7
    SKYBLUE2             = 111, 0x87afff
    CHARTREUSE2_1        = 112, 0x87d700
    DARKOLIVEGREEN3_1    = 113, 0x87d75f
    PALEGREEN3_1         = 114, 0x87d787
    DARKSEAGREEN3        = 115, 0x87d7af
    DARKSLATEGRAY3       = 116, 0x87d7d7
    SKYBLUE1             = 117, 0x87d7ff
    CHARTREUSE1          = 118, 0x87ff00
    LIGHTGREEN           = 119, 0x87ff5f
    LIGHTGREEN_1         = 120, 0x87ff87
    PALEGREEN1           = 121, 0x87ffaf
    AQUAMARINE1_1        = 122, 0x87ffd7
    DARKSLATEGRAY1       = 123, 0x87ffff
    RED3                 = 124, 0xaf0000
    DEEPPINK4_2          = 125, 0xaf005f
    MEDIUMVIOLETRED      = 126, 0xaf0087
    MAGENTA3             = 127, 0xaf00af
    DARKVIOLET_1         = 128, 0xaf00d7
    PURPLE_2             = 129, 0xaf00ff
    DARKORANGE3          = 130, 0xaf5f00
    INDIANRED            = 131, 0xaf5f5f
    HOTPINK3             = 132, 0xaf5f87
    MEDIUMORCHID3        = 133, 0xaf5faf
    MEDIUMORCHID         = 134, 0xaf5fd7
    MEDIUMPURPLE2        = 135, 0xaf5fff
    DARKGOLDENROD        = 136, 0xaf8700
    LIGHTSALMON3         = 137, 0xaf875f
    ROSYBROWN            = 138, 0xaf8787
    GREY63               = 139, 0xaf87af
    MEDIUMPURPLE2_1      = 140, 0xaf87d7
    MEDIUMPURPLE1        = 141, 0xaf87ff
    GOLD3                = 142, 0xafaf00
    DARKKHAKI            = 143, 0xafaf5f
    NAVAJOWHITE3         = 144, 0xafaf87
    GREY69               = 145, 0xafafaf
    LIGHTSTEELBLUE3      = 146, 0xafafd7
    LIGHTSTEELBLUE       = 147, 0xafafff
    YELLOW3              = 148, 0xafd700
    DARKOLIVEGREEN3_2    = 149, 0xafd75f
    DARKSEAGREEN3_1      = 150, 0xafd787
    DARKSEAGREEN2        = 151, 0xafd7af
    LIGHTCYAN3           = 152, 0xafd7d7
    LIGHTSKYBLUE1        = 153, 0xafd7ff
    GREENYELLOW          = 154, 0xafff00
    DARKOLIVEGREEN2      = 155, 0xafff5f
    PALEGREEN1_1         = 156, 0xafff87
    DARKSEAGREEN2_1      = 157, 0xafffaf
    DARKSEAGREEN1        = 158, 0xafffd7
    PALETURQUOISE1       = 159, 0xafffff
    RED3_1               = 160, 0xd70000
    DEEPPINK3            = 161, 0xd7005f
    DEEPPINK3_1          = 162, 0xd70087
    MAGENTA3_1           = 163, 0xd700af
    MAGENTA3_2           = 164, 0xd700d7
    MAGENTA2             = 165, 0xd700ff
    DARKORANGE3_1        = 166, 0xd75f00
    INDIANRED_1          = 167, 0xd75f5f
    HOTPINK3_1           = 168, 0xd75f87
    HOTPINK2             = 169, 0xd75faf
    ORCHID               = 170, 0xd75fd7
    MEDIUMORCHID1        = 171, 0xd75fff
    ORANGE3              = 172, 0xd78700
    LIGHTSALMON3_1       = 173, 0xd7875f
    LIGHTPINK3           = 174, 0xd78787
    PINK3                = 175, 0xd787af
    PLUM3                = 176, 0xd787d7
    VIOLET               = 177, 0xd787ff
    GOLD3_1              = 178, 0xd7af00
    LIGHTGOLDENROD3      = 179, 0xd7af5f
    TAN                  = 180, 0xd7af87
    MISTYROSE3           = 181, 0xd7afaf
    THISTLE3             = 182, 0xd7afd7
    PLUM2                = 183, 0xd7afff
    YELLOW3_1            = 184, 0xd7d700
    KHAKI3               = 185, 0xd7d75f
    LIGHTGOLDENROD2      = 186, 0xd7d787
    LIGHTYELLOW3         = 187, 0xd7d7af
    GREY84               = 188, 0xd7d7d7
    LIGHTSTEELBLUE1      = 189, 0xd7d7ff
    YELLOW2              = 190, 0xd7ff00
    DARKOLIVEGREEN1      = 191, 0xd7ff5f
    DARKOLIVEGREEN1_1    = 192, 0xd7ff87
    DARKSEAGREEN1_1      = 193, 0xd7ffaf
    HONEYDEW2            = 194, 0xd7ffd7
    LIGHTCYAN1           = 195, 0xd7ffff
    RED1                 = 196, 0xff0000
    DEEPPINK2            = 197, 0xff005f
    DEEPPINK1            = 198, 0xff0087
    DEEPPINK1_1          = 199, 0xff00af
    MAGENTA2_1           = 200, 0xff00d7
    MAGENTA1             = 201, 0xff00ff
    ORANGERED1           = 202, 0xff5f00
    INDIANRED1           = 203, 0xff5f5f
    INDIANRED1_1         = 204, 0xff5f87
    HOTPINK              = 205, 0xff5faf
    HOTPINK_1            = 206, 0xff5fd7
    MEDIUMORCHID1_1      = 207, 0xff5fff
    DARKORANGE           = 208, 0xff8700
    SALMON1              = 209, 0xff875f
    LIGHTCORAL           = 210, 0xff8787
    PALEVIOLETRED1       = 211, 0xff87af
    ORCHID2              = 212, 0xff87d7
    ORCHID1              = 213, 0xff87ff
    ORANGE1              = 214, 0xffaf00
    SANDYBROWN           = 215, 0xffaf5f
    LIGHTSALMON1         = 216, 0xffaf87
    LIGHTPINK1           = 217, 0xffafaf
    PINK1                = 218, 0xffafd7
    PLUM1                = 219, 0xffafff
    GOLD1                = 220, 0xffd700
    LIGHTGOLDENROD2_1    = 221, 0xffd75f
    LIGHTGOLDENROD2_2    = 222, 0xffd787
    NAVAJOWHITE1         = 223, 0xffd7af
    MISTYROSE1           = 224, 0xffd7d7
    THISTLE1             = 225, 0xffd7ff
    YELLOW1              = 226, 0xffff00
    LIGHTGOLDENROD1      = 227, 0xffff5f
    KHAKI1               = 228, 0xffff87
    WHEAT1               = 229, 0xffffaf
    CORNSILK1            = 230, 0xffffd7
    GREY100              = 231, 0xffffff
    GREY3                = 232, 0x080808
    GREY7                = 233, 0x121212
    GREY11               = 234, 0x1c1c1c
    GREY15               = 235, 0x262626
    GREY19               = 236, 0x303030
    GREY23               = 237, 0x3a3a3a
    GREY27               = 238, 0x444444
    GREY30               = 239, 0x4e4e4e
    GREY35               = 240, 0x585858
    GREY39               = 241, 0x626262
    GREY42               = 242, 0x6c6c6c
    GREY46               = 243, 0x767676
    GREY50               = 244, 0x808080
    GREY54               = 245, 0x8a8a8a
    GREY58               = 246, 0x949494
    GREY62               = 247, 0x9e9e9e
    GREY66               = 248, 0xa8a8a8
    GREY70               = 249, 0xb2b2b2
    GREY74               = 250, 0xbcbcbc
    GREY78               = 251, 0xc6c6c6
    GREY82               = 252, 0xd0d0d0
    GREY85               = 253, 0xdadada
    GREY89               = 254, 0xe4e4e4
    GREY93               = 255, 0xeeeeee


Fore256 = XtermFore()

# A collection of visually distinguishable colors, for use with colorizing identifiers
Fore256.differentiated_colors = (
    Fore256.BLUE,               # 0
    Fore256.CADETBLUE,          # 1
    Fore256.CHARTREUSE1,        # 2
    Fore256.CORNFLOWERBLUE,     # 3
    Fore256.CYAN1,              # 4
    Fore256.DARKGOLDENROD,      # 5
    Fore256.DARKORANGE,         # 6
    Fore256.DEEPPINK1,          # 7
    Fore256.GOLD1,              # 8
    Fore256.DARKKHAKI,          # 9
    Fore256.HONEYDEW2,          # 10
    Fore256.MEDIUMVIOLETRED,    # 11
    Fore256.PURPLE,             # 12
    Fore256.WHEAT1,             # 13
    Fore256.YELLOW2,            # 14
    Fore256.ROSYBROWN,          # 15
    Fore256.MAROON,             # 16
    Fore256.LIGHTCYAN3,         # 17
    Fore256.HOTPINK_1,          # 18
    Fore256.GREY82,             # 19
    Fore256.LIGHTCORAL,         # 20
    Fore256.LIGHTSEAGREEN,      # 21
    Fore256.MEDIUMSPRINGGREEN,  # 22
    Fore256.OLIVE,              # 23
    Fore256.DODGERBLUE2,        # 24
    Fore256.ORANGERED1,         # 25
    Fore256.PALETURQUOISE1,     # 26
    Fore256.THISTLE3,           # 27
    Fore256.DARKTURQUOISE,      # 28
    Fore256.GREEN,              # 29
    Fore256.LIGHTGOLDENROD1,    # 30
    Fore256.LIGHTSALMON1,       # 31
    Fore256.PINK1,              # 32
    Fore256.NAVAJOWHITE1,       # 33
    Fore256.LIGHTSLATEBLUE,     # 34
    Fore256.LIGHTCYAN1,         # 35
    Fore256.GOLD3_1,            # 36
    Fore256.INDIANRED,          # 37
    Fore256.PURPLE_2,           # 38
    Fore256.SALMON1,            # 39
)


def print_colors(colors: Iterable[XtermColor]):
    """Print swatch, code, and name for each of the specified colours
    """
    for color in colors:
        print(color.swatch, color(f'{color.code:>4d} {color.name}'))


def print_all_colors():
    """Print swatch, code, and name for all xterm-256 colours
    """
    print_colors(Fore256.all_colors.values())


def print_differentiated_colors():
    """Print swatch, code, and name for each of the curated "differentiated" colours
    """
    print_colors(Fore256.differentiated_colors)


def print_color_table(colors: Iterable[XtermColor], sort: bool = True):
    if sort:
        colors = sorted(colors, key=lambda color: color.code)
    elif not isinstance(colors, Sequence):
        colors = tuple(colors)

    @contextmanager
    def row(title: Any = '', *headers):
        print_cell(title)
        for cell in headers:
            print_cell(cell)

        yield
        print()

    def th():
        return row('', '')

    def print_cell(s: Any = '') -> None:
        print(f'{s:^4}', end='')

    def swatch(color: XtermColor, n: int, c: str = '▇') -> str:
        return color.swatch(n=n, c=c)

    def print_comparison(a: XtermColor, b: XtermColor) -> None:
        c = '▇'
        print_cell(f' {swatch(a, 1)}{swatch(b, 1)} ')

    # Header
    with th():
        for color in colors:
            print_cell(swatch(color, 4))

    with th():
        for color in colors:
            print_cell(color.code)

    # Body
    for row_color in colors:
        with row(swatch(row_color, 4), row_color.code):
            for col_color in colors:
                if row_color is not col_color:
                    print_comparison(row_color, col_color)
                else:
                    print_cell()


ColorDistanceMatrix = Dict[XtermColor, Dict[XtermColor, float]]


def calculate_distance_matrix(colors: Iterable[XtermColor]) -> ColorDistanceMatrix:
    """Calculate the Delta-E 2000 colour distance between each of the specified colours

    Note that both the forward and backward directions are stored in the resulting
    dictionary — i.e. dist_mat[RED][YELLOW] == dist_mat[YELLOW][RED]
    """
    _require_colormath()
    from colormath.color_diff import delta_e_cie2000

    if not isinstance(colors, Sequence):
        colors = tuple(colors)

    dist_mat = defaultdict(dict)
    for i, a in enumerate(colors):
        for b in colors[i+1:]:
            dist = delta_e_cie2000(a.as_lab_color, b.as_lab_color)
            dist_mat[a][b] = dist_mat[b][a] = dist

    return dist_mat


def find_differentiated_colors(colors: Iterable[XtermColor],
                               n: int,
                               dist_mat: ColorDistanceMatrix = None,
                               min_dist: float = float('-Inf'),
                               ) -> Set[XtermColor]:
    """Find N colors most(-ish) different from each other

    :param colors:
        Population of colours to select differentiated subset from.

    :param n:
        Number of colours to include in returned subset.

    :param dist_mat:
        A pre-calculated colour distance matrix to use. If not supplied, one
        will be calculated.

    :param min_dist:
        Minimum distance each colour must have between each other in the subset.

    """
    if not isinstance(colors, Sequence):
        colors = tuple(colors)

    if dist_mat is None:
        dist_mat = calculate_distance_matrix(colors)

    best_pair = ()
    max_dist = 0

    for i, a in enumerate(colors):
        for b in colors[i+1:]:
            if (dist := dist_mat[a][b]) > max_dist:
                max_dist = dist
                best_pair = (a, b)

    subset = {*best_pair}

    while len(subset) < n:
        max_avg_dist = 0
        v_avg_best = None

        for v in colors:
            if v in subset:
                continue

            dists = []
            is_below_min_dist = False
            for v_prime in subset:
                dist = dist_mat[v][v_prime]
                if dist < min_dist:
                    is_below_min_dist = True
                    break

                dists.append(dist)

            if is_below_min_dist:
                continue

            if (avg_dist := statistics.mean(dists)) > max_avg_dist:
                max_avg_dist = avg_dist
                v_avg_best = v

        subset.add(v_avg_best)

    return subset



import asyncio
import inspect
import operator
import os
import re
from contextlib import contextmanager, nullcontext
from itertools import chain, zip_longest
from typing import (
    Any,
    Awaitable,
    Callable,
    Collection,
    Coroutine,
    Dict,
    Iterable,
    List,
    Optional,
    TYPE_CHECKING,
    Union,
)


from asciimatics import screen
from asciimatics.event import KeyboardEvent
from asciimatics.exceptions import ResizeScreenError
from asciimatics.screen import Canvas, Screen
from asciimatics.widgets import Label, MultiColumnListBox
from terminaltables.width_and_alignment import RE_COLOR_ANSI
from wcwidth import wcwidth

from tui_progress.itertools import chunk
from tui_progress.terminal import original_stdstreams, visible_width

if TYPE_CHECKING:
    from tui_progress.table import RowDefn, SingleTable


__all__ = [
    'open_screen',
    'ManagedTui',
    'ColoredLabel',
    'ColoredMultiColumnListBox',
]


def open_screen(*args, **kwargs) -> Screen:
    """Open an asciimatics screen, ensuring size is set properly

    Sometimes ncurses doesn't report the correct terminal size,
    causing window resizes to leave a small TUI in a large window;
    or worse: cause wrapping to occur, absolutely mutilating the TUI.

    This method grabs the proper terminal size ourselves and reports it
    to the asciimatics screen.
    """
    size = os.get_terminal_size()

    kwargs['height'] = size.lines
    screen = Screen.open(*args, **kwargs)
    screen.height, screen.width = size.lines, size.columns

    return screen


class ManagedScreen(screen.ManagedScreen):
    """asciimatic's ManagedScreen, ensuring window resizes are properly handled"""

    def _open_screen(self) -> Screen:
        return open_screen()

    def __call__(self, *args, **kwargs):
        screen = self._open_screen()
        kwargs["screen"] = screen
        output = self.func(*args, **kwargs)
        screen.close()
        return output

    def __enter__(self):
        """
        Method used for with statement
        """
        self.screen = self._open_screen()
        return self.screen


class ManagedTui:
    """Renders and manages an async asciimatics Screen"""

    def __init__(
        self,
        init_screen: Callable[[Screen], Awaitable],
        tasks: Iterable[Union[asyncio.Task, Coroutine, Callable[[], Coroutine]]],
        *,
        tick: float = 0.05,
        enable_pdb: bool = False,
    ):
        self.init_screen = init_screen
        self.tasks = tasks
        self.tick = tick
        self.enable_pdb = enable_pdb

        self.screen: Optional[Screen] = None
        self._managed_tasks: Optional[List[asyncio.Task]]

    async def __aenter__(self):
        self._managed_tasks = []

        tasks = [
            self._managed_screen_loop,
            *self.tasks,
        ]
        for task in tasks:
            if inspect.iscoroutinefunction(task):
                task = task()
            if inspect.iscoroutine(task):
                task = asyncio.create_task(task)

            if isinstance(task, asyncio.Task):
                self._managed_tasks.append(task)
            else:
                raise TypeError(
                    f'Provided tasks must be asyncio.Tasks, '
                    f'coroutine functions (async def), or coroutines. '
                    f'Found {type(task)} for {task!r}')

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        for task in self._managed_tasks:
            task.cancel()

        if self.screen:
            self.screen.close()

    async def force_update(self):
        if self.screen:
            self.screen.force_update()

    async def _start_screen(self):
        self.screen = open_screen()
        await self.init_screen(self.screen)

    async def _managed_screen_loop(self):
        with original_stdstreams():
            while True:
                await self._start_screen()
                try:
                    await self._screen_render_loop()
                except ResizeScreenError:
                    # Close the screen and continue the loop, where a new Screen
                    # will be initialized with the new window size.
                    self.screen.close()

    async def _screen_render_loop(self):
        with (self.breakpoint_on_error() if self.enable_pdb else nullcontext()):
            while True:
                self.screen.draw_next_frame()
                if self.screen.has_resized():
                    raise ResizeScreenError('Screen resized')
                await asyncio.sleep(self.tick)

    @contextmanager
    def breakpoint_on_error(self):
        """Debugging helper — on Exception: closes screen, prints traceback, drops to PDB
        """
        import traceback

        try:
            yield
        except (ResizeScreenError, KeyboardInterrupt):
            raise
        except Exception:
            if self.screen:
                self.screen.close()
            traceback.print_exc()
            breakpoint()
            raise

    async def breakpoint(self, reopen_screen: bool = True):
        """Debugging helper — closes screen, prints traceback, drops to PDB
        """
        if self.screen:
            self.screen.close()
        else:
            reopen_screen = False

        breakpoint()

        if reopen_screen:
            self.screen = Screen.open()
            await self.init_screen(self.screen)


class ColoredMultiColumnListBox(MultiColumnListBox):
    ALIGNMENTS = {
        'left': '<',
        'center': '^',
        'right': '>',
    }

    def __init__(self,
                 *args,
                 key: Union[str, Collection[str], Callable[['RowDefn'], Any]] = None,
                 **kwargs):
        """A MultiColumnListBox supporting the output of colours (and other invisible chars)

        :param key:
            A key in each row that uniquely identifies it, or a function that accepts
            the row and returns a unique identifier. This will be used to ensure user
            selections are preserved across updates (if possible).

        """
        if not isinstance(key, str) and isinstance(key, Collection):
            keys = tuple(key)
            if len(keys) == 1:
                key, = keys
            else:
                key = lambda row: tuple(row[k] for k in keys)

        if isinstance(key, str):
            key = operator.itemgetter(key)

        self.key = key

        super().__init__(*args, **kwargs)

    @property
    def columns(self):
        return self._columns

    @columns.setter
    def columns(self, columns):
        self._columns = []
        self._align = []
        self._spacing = []
        for i, column in enumerate(columns):
            if isinstance(column, int):
                self._columns.append(column)
                self._align.append("<")
            else:
                match = re.match(r"([<>^]?)(\d+)([%]?)", column)
                self._columns.append(float(match.group(2)) / 100
                                     if match.group(3) else int(match.group(2)))
                self._align.append(match.group(1) if match.group(1) else "<")
            self._spacing.append(1 if (i > 0 and self._align[i] == "<" and self._align[i-1] == ">") else 0)

    @property
    def titles(self):
        return self._titles

    @titles.setter
    def titles(self, titles):
        self._titles = titles

    @property
    def height(self):
        return self._required_height

    @height.setter
    def height(self, height):
        self._required_height = height

    def parse_attrs_from_table(self, table: 'SingleTable') -> Dict[str, Any]:
        return dict(
            columns=[
                f'{self.ALIGNMENTS.get(justify, "")}{width + 2}'
                for justify, width in zip(table.justify_columns.values(), table.column_widths)
            ],
            options=[
                (row, self.key(table.source_dicts[i]) if self.key else i)
                for i, row in enumerate(table.table_data[1:])
            ],
            titles=table.table_data[0],
        )

    def update_from_attrs(self, attrs: Dict[str, Any], preserve_value: bool = True):
        with self.value_preserved() if preserve_value else nullcontext():
            for name, value in attrs.items():
                setattr(self, name, value)

    def update_from_table(self, table: 'SingleTable'):
        attrs = self.parse_attrs_from_table(table)
        self.update_from_attrs(attrs)

    @contextmanager
    def value_preserved(self):
        last_value = self.value
        yield
        self.value = last_value

    def process_event(self, event):
        if isinstance(event, KeyboardEvent):
            if len(self._options) > 0 and event.key_code == Screen.KEY_HOME:
                self._line = 0
                self.value = self._options[self._line][1]
            elif len(self._options) > 0 and event.key_code == Screen.KEY_END:
                self._line = len(self._options) - 1
                self.value = self._options[self._line][1]
            else:
                return super().process_event(event)
        else:
            return super().process_event(event)

    def update(self, frame_no):
        self._draw_label()

        # Calculate new visible limits if needed.
        height = self._h
        width = self._w
        dy = 0

        # Clear out the existing box content
        (colour, attr, bg) = self._frame.palette["field"]
        for i in range(height):
            self._frame.canvas.print_at(
                " " * width,
                self._x + self._offset,
                self._y + i + dy,
                colour, attr, bg)

        # Allow space for titles if needed.
        if self._titles:
            dy += 1
            height -= 1

        # Decide whether we need to show or hide the scroll bar and adjust width accordingly.
        if self._add_scroll_bar:
            self._add_or_remove_scrollbar(width, height, dy)
        if self._scroll_bar:
            width -= 1

        # Now draw the titles if needed.
        if self._titles:
            row_dx = 0
            colour, attr, bg = self._frame.palette["title"]
            for i, (title, align, space) in enumerate(zip(self._titles, self._align, self._spacing)):
                cell_width = self._get_width(self._columns[i], width)
                vis_width = visible_width(title)

                left_pad = 0
                right_pad = 0
                pad = cell_width - vis_width
                if align == '^':
                    left_pad = pad // 2
                    right_pad = pad - left_pad
                elif align == '>':
                    left_pad = pad
                else:
                    right_pad = pad

                _print_at_color_aware(
                    self._frame.canvas,
                    "{}{}{}".format(
                        " " * (space + left_pad),
                        title,
                        " " * right_pad,
                        ),
                    self._x + self._offset + row_dx,
                    self._y,
                    colour, attr, bg,
                    )
                row_dx += cell_width + space

        # Don't bother with anything else if there are no options to render.
        if len(self._options) <= 0:
            return

        # Render visible portion of the text.
        self._start_line = max(0, max(self._line - height + 1,
                                      min(self._start_line, self._line)))
        for i, [row, _] in enumerate(self._options):
            if not (self._start_line <= i < self._start_line + height):
                continue

            colour, attr, bg = self._pick_colours("field", i == self._line)
            row_dx = 0
            # Try to handle badly formatted data, where row lists don't
            # match the expected number of columns.
            for text, cell_width, align, space in zip_longest(row, self._columns, self._align, self._spacing, fillvalue=""):
                if cell_width == "":
                    break
                cell_width = self._get_width(cell_width, width)
                vis_width = visible_width(str(text))

                left_pad = 0
                right_pad = 0
                pad = cell_width - vis_width
                if align == '^':
                    left_pad = pad // 2
                    right_pad = pad - left_pad
                elif align == '>':
                    left_pad = pad
                else:
                    right_pad = pad

                _print_at_color_aware(
                    self._frame.canvas,
                    "{}{}{}".format(
                        " " * (space + left_pad),
                        text,
                        " " * right_pad,
                        ),
                    self._x + self._offset + row_dx,
                    self._y + i + dy - self._start_line,
                    colour, attr, bg,
                    )
                row_dx += cell_width + space

        # And finally draw any scroll bar.
        if self._scroll_bar:
            self._scroll_bar.update()


class ColoredLabel(Label):
    """Label widget supporting ANSI/xterm-256 colour printing
    """
    def update(self, frame_no):
        palette_key = self._pick_palette_key('label', selected=False, allow_input_state=False)
        color, attr, bg = self._frame.palette[palette_key]
        _print_at_color_aware(
            self._frame.canvas,
            self._text,
            self._x, self._y,
            color, attr, bg,
        )


def _print_at_color_aware(canvas: Canvas, text, x, y, color, attr, bg):
    RE_COLOR_ANSI = re.compile(r'(\033\[([\d;]+)m)', re.UNICODE)

    parts = RE_COLOR_ANSI.split(text)

    for escape, code, chars in chunk(chain(['\x1b[0m', str(color)], parts), 3):
        if not chars:
            continue

        ops = [int(op) for op in code.split(';')]
        if len(ops) == 1:
            color = ops[0]
        elif len(ops) == 3 and ops[:2] == [38, 5]:
            color = ops[2]

        canvas.print_at(chars, x, y, color, attr, bg)
        x += len(chars)


def _enforce_width_color_aware(text, width):
    # Double-width strings cannot be more than twice the string length, so no need to try
    # expensive truncation if this upper bound isn't an issue.
    if 2 * len(text) < width:
        return text

    final = ''
    parts = RE_COLOR_ANSI.split(text)

    size = 0
    for code, part in chunk(chain([''], parts), 2):
        final += code

        if size >= width:
            continue
        else:
            for i, c in enumerate(part):
                w = wcwidth(c) if ord(c) >= 256 else 1
                if size + w > width:
                    final += part[:i]
                    break
                size += w
            else:
                final += part

    return final



from typing import Dict, Generic, Hashable, Iterable, Iterator, TypeVar

__all__ = ['OrderedSet']

V = TypeVar('V', bound=Hashable)


class OrderedSet(Generic[V]):
    """Bare implementation of set, preserving insertion order"""

    def __init__(self, items: Iterable[V] = ()):
        self._items: Dict[V, None] = {item: None for item in items}

    def __iter__(self) -> Iterator[V]:
        return iter(self._items)

    def __contains__(self, item: V):
        return item in self._items

    def add(self, item: V):
        self._items[item] = None

    def remove(self, item: V):
        del self._items[item]

    def discard(self, item: V):
        try:
            self.remove(item)
        except KeyError:
            pass

    def clear(self):
        self._items.clear()

    def update(self, items: Iterable[V]):
        self._items.update({item: None for item in items})



from .color import Fore256
from .table import (
    render_table,
    display_table,
    watch_table,
    WatchTable,
    AsyncTable,
)
from .terminal import (
    NestedHalo,
    AppendingHalo,
    TimedHalo,
    new_subtask,
    timed_subtask,
    nested_tqdm,
    subtqdm,
    subtrange,
)
from .tui import (
    ManagedTui,
    ColoredMultiColumnListBox,
    ColoredLabel,
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tui_progress']

package_data = \
{'': ['*']}

install_requires = \
['asciimatics>=1.13.0,<2.0.0',
 'cached-property>=1.5.2,<2.0.0',
 'cursor>=1.3.4,<2.0.0',
 'halo>=0.0.31,<0.0.32',
 'humanize>=4.0.0,<5.0.0',
 'log-symbols>=0.0.14,<0.0.15',
 'terminaltables>=3.1.10,<4.0.0',
 'tqdm>=4.63.0,<5.0.0',
 'wcwidth>=0.2.5,<0.3.0',
 'wrapt>=1.14.0,<2.0.0']

setup_kwargs = {
    'name': 'tui-progress',
    'version': '0.1.1',
    'description': '',
    'long_description': None,
    'author': 'Zach Kanzler',
    'author_email': 'they4kman@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import functools
import inspect
import os
import sys
import unicodedata
from contextlib import ContextDecorator, contextmanager
from datetime import datetime
from typing import Callable, ContextManager, Optional, TextIO, Union

import wrapt
from halo import Halo
from halo._utils import decode_utf_8_text
from humanize import naturaldelta
from log_symbols.symbols import LogSymbols
from terminaltables.width_and_alignment import RE_COLOR_ANSI
from tqdm import tqdm

from tui_progress.progress import ProgressIndicator

TERM_MOVE_UP = '\x1b[A'
TERM_CLEAR_LINE = '\x1b[2K'

NOTSET = object()


class SubtaskComplete(Exception):
    """Non-error signal exception to print subtask completion message

    If this is fired, the subtask knows not to print its own "Done"
    """
    def __init__(self, message: str = None):
        self.message = message


class Subtask(wrapt.ObjectProxy):
    """tqdm wrapper with extra subtask methods
    """
    step_prefix: str

    def __init__(self, wrapped, *, step_prefix: str):
        super().__init__(wrapped)
        self.step_prefix = step_prefix

    def done(self, msg: Optional[str] = 'Done!'):
        """Stop subtask execution and optionally print message"""
        raise SubtaskComplete(msg)

    def progress(self, *args, **kwargs) -> tqdm:
        default_bar_format = (
            '[{n_fmt} / {total_fmt}] '
            '{desc}{bar} '
            '[{elapsed}<{remaining}, {rate_fmt}{postfix}]'
        )
        bar_format = kwargs.pop('bar_format', None) or default_bar_format
        kwargs['bar_format'] = f'{self.step_prefix}{bar_format}'

        kwargs.setdefault('position', (self.pos or 0) + 1)

        pbar = tqdm(*args, **kwargs)
        return pbar


class NestedHalo(Halo):
    """Halo spinner which may be printed above/below other spinners
    """
    def __init__(self, text='', color='cyan', spinner=None,
                 animation=None, placement='left', interval=-1, enabled=True, stream=sys.stderr):
        super().__init__(text=text, color=color, spinner=spinner, animation=animation,
                         placement=placement, interval=interval, enabled=enabled, stream=stream,)

        self.indicator = ProgressIndicator(self._stream, name=self.__class__.__name__)

    def start(self, text=None):
        self.indicator.update_loop(self.frame, 0.001 * self._interval)
        return self

    def stop(self):
        self.indicator.stop()
        return self

    def stop_and_persist(self, symbol=' ', text=None):
        symbol = decode_utf_8_text(symbol)

        if text is not None:
            text = decode_utf_8_text(text)
        else:
            text = self._text['original']
        text = text.strip()

        if self._placement == 'right':
            prefix, suffix = text, symbol
        else:
            prefix, suffix = symbol, text

        output = f'{prefix} {suffix}'
        self.indicator.complete(output)

    def __call__(self, f):
        """Use Halo as decorator (may be used with async functions)"""

        if inspect.iscoroutinefunction(f):
            @functools.wraps(f)
            async def wrapped(*args, **kwargs):
                with self:
                    return await f(*args, **kwargs)

        else:
            @functools.wraps(f)
            def wrapped(*args, **kwargs):
                with self:
                    return f(*args, **kwargs)

        return wrapped


class AppendingHalo(NestedHalo):
    """Halo spinner whose message methods append to current text, not overwrite
    """
    def __init__(self, text='', color='cyan', spinner=None, animation=None,
                 placement='left', interval=-1, enabled=True, stream=sys.stderr, **kwargs):
        self.original_text = text
        self.printed_indicators: list[ProgressIndicator] = []

        super().__init__(text, color, spinner, animation, placement, interval,
                         enabled, stream, **kwargs)

    def __exit__(self, type, value, traceback):
        # Don't return a value, because the exception is silenced otherwise.
        # (The original Halo class does this)
        super().__exit__(type, value, traceback)

    def stop(self):
        for indicator in self.printed_indicators:
            indicator.complete(indicator.frame)
        return super().stop()

    def _append_message(self, method, text=None, overwrite=False):
        if not overwrite:
            text = f'{self.original_text} {text}'
        return method(text)

    def succeed(self, text=None, overwrite=False):
        return self._append_message(super().succeed, text=text, overwrite=overwrite)

    def fail(self, text=None, overwrite=False):
        return self._append_message(super().fail, text=text, overwrite=overwrite)

    def warn(self, text=None, overwrite=False):
        return self._append_message(super().warn, text=text, overwrite=overwrite)

    def info(self, text=None, overwrite=False):
        return self._append_message(super().info, text=text, overwrite=overwrite)

    def print(self, text: str = ''):
        """Print a message above the Halo line
        """
        indicator = ProgressIndicator(self._stream)
        indicator.start()
        indicator.update(text)
        self.printed_indicators.append(indicator)

    def _print_with_symbol(self, symbol: str, text: str):
        self.print(f'{symbol} {text}')

    def print_succeed(self, text: str):
        self._print_with_symbol(LogSymbols.SUCCESS.value, text)

    def print_fail(self, text: str):
        self._print_with_symbol(LogSymbols.ERROR.value, text)

    def print_warn(self, text: str):
        self._print_with_symbol(LogSymbols.WARNING.value, text)

    def print_info(self, text: str):
        self._print_with_symbol(LogSymbols.INFO.value, text)


class TimedHalo(AppendingHalo):
    """Halo spinner showing elapsed time when succeed(), fail(), etc called
    """
    start_dt: Optional[datetime]
    end_dt: Optional[datetime]
    elapsed: Union[bool, str, Callable[[datetime, datetime], str]]

    def __init__(
        self,
        *args,
        elapsed: Union[bool, str, Callable[[datetime, datetime], str]] = True,
        **kwargs,
    ):
        """
        :param elapsed:
            Determines whether/how to show elapsed time with printed messages.
            If False, no elapsed times will be displayed with messages.
            If True, the elapsed time will be displayed as "[Δ 2 seconds]".
            If a string, it will be printed verbatim.
            If a callable, the passed method will be called with (start_dt, end_dt),
            and its return value will be printed.
        """
        super().__init__(*args, **kwargs)
        self.start_dt = None
        self.end_dt = None
        self.elapsed = elapsed

    def start_timer(self):
        self.start_dt = datetime.utcnow()

    def stop_timer(self):
        if not self.end_dt:
            self.end_dt = datetime.utcnow()

    def start(self, text=None):
        self.start_timer()
        return super().start(text)

    def stop(self):
        self.stop_timer()
        return super().stop()

    def stop_and_persist(self, symbol=' ', text=None):
        self.stop_timer()

        if text is not None:
            text = decode_utf_8_text(text)
        else:
            text = self._text['original']

        elapsed_text = self.format_elapsed()
        if elapsed_text:
            text = f'{text} {elapsed_text}'

        return super().stop_and_persist(symbol, text)

    def format_elapsed(self) -> str:
        """Return a string representation of the elapsed time
        """
        if not self.elapsed:
            return ''
        elif isinstance(self.elapsed, str):
            return self.elapsed
        elif callable(self.elapsed):
            return self.elapsed(self.start_dt, self.end_dt)
        else:
            elapsed = self.end_dt - self.start_dt
            return f' [Δ {naturaldelta(elapsed)}]'


devnull = open(os.devnull, 'w')


@contextmanager
def new_subtask(
    message: str,
    finished: Optional[str] = 'Done!',
    *,
    show_elapsed: bool = False,
    enabled: bool = True,
    **halo_kwargs,
) -> Union[ContextManager[TimedHalo], ContextDecorator]:
    """Show a spinner during part of a task, appending message upon completion

    NOTE: an ellipsis ("...") will be appended to the message.
    """
    stream = sys.stderr if enabled else devnull

    with TimedHalo(f'{message} ...', stream=stream, elapsed=show_elapsed, **halo_kwargs) as subtask:
        try:
            yield subtask
        except Exception:
            subtask.fail('ERROR')
            raise
        else:
            if finished is not None:
                subtask.succeed(finished)


def timed_subtask(*args, **kwargs) -> Union[ContextManager[TimedHalo], ContextDecorator]:
    """Shortcut for new_subtask(..., show_elapsed=True)"""
    return new_subtask(*args, **kwargs, show_elapsed=True)


class nested_tqdm(tqdm):
    """tqdm which may be nested/run concurrently with other nested_tqdm's/NestedHalos
    """
    def __init__(self, *args, file: TextIO = sys.stderr, **kwargs):
        self.indicator = ProgressIndicator(file, name=self.__class__.__name__)
        self.indicator.start()
        super().__init__(*args, file=file, **kwargs)

    def close(self):
        if self.leave:
            self.ncols = 0  # only show summary
            self.indicator.complete(self.__str__())
        else:
            self.indicator.stop()

        self.disable = True

    def status_printer(self, file):
        return self.indicator.update

    def moveto(self, n):
        pass  # noop


class subtqdm(nested_tqdm):
    """tqdm with subtask convenience contextmanager
    """
    @contextmanager
    def subtask(self,
                title: str = None,
                default_done_msg: Optional[str] = 'Done!',
                print_spacer: bool = True,
                include_step_prefix: bool = True,
                ) -> ContextManager['Subtask']:
        """Declare a subtask, which updates the task pbar upon completion
        """
        step_prefix = ''
        if include_step_prefix:
            step_number = self.n + 1
            step_prefix = f'[STEP {step_number}] '

        if title is not None:
            self.write(f'{step_prefix}{title}')

        try:
            yield Subtask(self, step_prefix=step_prefix)
        except SubtaskComplete as done:
            message = done.message
        else:
            message = default_done_msg

        if message:
            self.write(f'{step_prefix}{message}')
        if print_spacer:
            self.write('')

        self.update()


class subtrange(subtqdm):
    """A shortcut for subtqdm(xrange(*args), **kwargs).
    """
    def __init__(self, *args, **kwargs):
        iterable = range(*args)
        super().__init__(iterable, **kwargs)


@contextmanager
def original_stdstreams():
    """Restore native stdout/stderr streams temporarily

    Upon import, the halo library automatically calls colorama.init(), which
    replaces stdout & stderr with wrapper streams, interfering with color output
    in asciimatics. Unfortunately, calling colorama.deinit() has *not* resolved
    the issue — stdout & stderr remain wrappers — but restoring stdout and stderr
    with their original values (provided by Python as sys.__stdout__ and
    sys.__stderr__) does resolve the issue.
    """
    prev_stdout = sys.stdout
    prev_stderr = sys.stderr

    try:
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__
        yield
    finally:
        sys.stdout = prev_stdout
        sys.stderr = prev_stderr


def visible_length(unistr):
    """Returns the number of printed characters in a Unicode string.

    Source: https://stackoverflow.com/a/30888594/148585
    """
    return sum(1 for char in unistr if unicodedata.combining(char) == 0)


def visible_width(string):
    """Get the visible width of a unicode string.

    This version of terminaltables's visible_width also takes into account
    Unicode characters which provide ligatures or styling, which do not
    have any width.
    """
    if '\033' in string:
        string = RE_COLOR_ANSI.sub('', string)

    # Convert to unicode.
    try:
        string = string.decode('u8')
    except (AttributeError, UnicodeEncodeError):
        pass

    return visible_length(string)



import asyncio
import inspect
import sys
import time
import traceback
from contextlib import suppress
from datetime import datetime
from functools import cmp_to_key
from threading import Event, Thread
from typing import (
    Any, AsyncIterator, Awaitable, Callable, Collection, Dict, IO, Iterable,
    List, Literal, Optional, Tuple, TypedDict, Union,
)

import terminaltables
from asciimatics.exceptions import NextScene, ResizeScreenError, StopApplication
from asciimatics.scene import Scene
from asciimatics.screen import Screen
from asciimatics.widgets import Frame, Layout, Widget
from terminaltables.build import flatten
from terminaltables.width_and_alignment import RE_COLOR_ANSI

from tui_progress.terminal import original_stdstreams, visible_width
from tui_progress.tui import ColoredMultiColumnListBox, ManagedScreen

ColumnDefnSource = Union[str, 'FullColumnDefn', 'ColumnDefinition']
ValueRow = Dict[str, Any]
SeparatorRow = None
RowDefn = Union[ValueRow, SeparatorRow]

ValueTransformer = Callable[[Any], Any]
RowValueTransformer = Callable[[Any, Dict[str, Any], str, Dict[str, 'ColumnDefinition']], Any]
TransformerFunc = Union[ValueTransformer, RowValueTransformer]
TransformerSource = Union[str, TransformerFunc, Collection[TransformerFunc]]

InputsSource = Union[Iterable[RowDefn], Callable[[], Iterable[RowDefn]]]


class FullColumnDefn(TypedDict, total=False):
    title: str
    is_key: bool
    formatter: TransformerSource
    justify: Literal['left', 'right', 'center']
    sort: Optional[Literal['asc', 'desc']]
    sort_order: Union[float, int]
    sort_key: TransformerSource
    min_width: Optional[int]


class ColumnDefinition:
    # Key in row data to find the column value
    name: str

    # Title to display for the column
    title: str

    # Whether the column value should be used to uniquely identify the row
    #  (used in interactive tables, to ensure selection persists across updates)
    is_key: bool

    # Minimum width of the column
    min_width: Optional[int]

    # Transform a column value to a displayable format
    formatter: RowValueTransformer
    source_formatter: TransformerSource

    # Which direction to align column values to
    justify: Literal['left', 'right', 'center']

    # How to sort the column: asc or desc
    sort: Optional[Literal['asc', 'desc']]

    # If multiple columns are sorted, in which order is this column sorted
    sort_order: Union[float, int]

    # Transform a column into a sortable value
    sort_key: TransformerSource

    # Extra user-supplied keys to provide metadata on the column
    meta: Dict[str, Any]

    def __init__(
        self,
        name: str,
        title: str = None,
        is_key: bool = False,
        min_width: int = None,
        formatter: TransformerSource = lambda s: s,
        justify: Literal['left', 'right', 'center'] = 'left',
        sort: Optional[Literal['asc', 'desc']] = None,
        sort_order: Union[float, int] = float('Inf'),
        sort_key: TransformerSource = lambda s: s,
        **meta
    ):
        self.name = name
        self.title = title if title is not None else name.capitalize()
        self.is_key = is_key
        self.min_width = min_width

        self.source_formatter = formatter
        self.formatter = init_transformer(self.source_formatter)
        self.justify = justify

        self.sort = sort.lower() if sort else None
        self.sort_order = sort_order
        self.sort_key = init_transformer(sort_key)

        self.meta = meta

        self._cmp = CMP_STRATEGIES.get(self.sort)

    @property
    def is_sortable(self):
        return self._cmp is not None

    @property
    def cmp(self):
        return self._cmp

    def get_title(self) -> str:
        title = self.title

        if self.min_width:
            justify = {
                'left': str.ljust,
                'center': str.center,
                'right': str.rjust,
            }

            lines = title.splitlines()
            lines[0] = justify[self.justify](lines[0], self.min_width)
            title = '\n'.join(lines)

        return title

    def get_row_sort_key(self, data: Dict[str, Any], columns: Dict[str, 'ColumnDefinition']) -> Any:
        value = data[self.name]
        return self.cmp(self.sort_key(value, data, self.name, columns))

    def get_row_formatted_value(self, data: Dict[str, Any], columns: Dict[str, 'ColumnDefinition']) -> Any:
        value = data[self.name]
        return self.formatter(value, data, self.name, columns)


CMP_STRATEGIES = {
    'asc' : cmp_to_key(lambda x, y: (x > y) - (x < y)),
    'desc': cmp_to_key(lambda x, y: (x < y) - (x > y)),
}


def is_row_value_transformer(transformer: TransformerFunc) -> bool:
    """Determine whether a transformer func accepts row args

    >>> is_row_value_transformer(lambda v: v)
    False
    >>> is_row_value_transformer(lambda value, data, column, columns: value)
    True
    """
    sig = inspect.signature(transformer)

    try:
        sig.bind('value', 'data', 'column', 'columns')
    except TypeError:
        return False
    else:
        return True


def as_row_value_transformer(transformer: ValueTransformer) -> RowValueTransformer:
    """Wrap a value-transformer to be used as a row value-transformer
    """
    return lambda value, data, column, columns: transformer(value)


def init_transformer(transformer: TransformerSource) -> RowValueTransformer:
    """Process any type of transformer source into a row value-transformer
    """
    ###
    # A transformer may be the name of another column
    #
    if isinstance(transformer, str):
        column_name = transformer

        def get_column(value, data, column, columns):
            return data[column_name]

        get_column.__name__ = f'get_{column_name}'
        return get_column

    ###
    # A transformer may be a function that accepts either a single column value,
    # or the whole row's data + metadata (column value, row data, name of column,
    # and column definition)
    #
    elif callable(transformer):
        if not is_row_value_transformer(transformer):
            transformer = as_row_value_transformer(transformer)
        return transformer

    ###
    # A transformer may be a collection of the above types, which will be called
    # in sequence, using the previous transformer's return value as the next
    # transformer's column value (the only argument if a value-transformer, or
    # the first argument if a row value-transformer)
    #
    elif isinstance(transformer, Collection):
        transformers = transformer
        transformers = [
            init_transformer(transformer)
            for transformer in transformers
        ]

        def chained_transformer(value, data, column, columns):
            for transformer in transformers:
                value = transformer(value, data, column, columns)
            return value

        return chained_transformer

    else:
        raise ValueError(
            f'A transformer must either be a string, a single-argument method, '
            f'a four-argument method, or a list of all three. '
            f'Found: ({type(transformer)}) {transformer!r}')


def init_column(name: str,
                defn: ColumnDefnSource,
                **defaults,
                ) -> ColumnDefinition:
    """Fill a column's definition with defaults
    """
    if isinstance(defn, ColumnDefinition):
        return defn

    if isinstance(defn, str):
        defn = {
            'title': defn,
        }

    return ColumnDefinition(name, **defaults, **defn)


def init_columns(columns: Dict[str, ColumnDefnSource],
                 **defaults,
                 ) -> Dict[str, ColumnDefinition]:
    """Fill all column definitions with defaults
    """
    return {
        column: init_column(column, defn, **defaults)
        for column, defn in columns.items()
    }


def sort_by_columns(columns: Dict[str, ColumnDefinition],
                    inputs: Iterable[RowDefn],
                    ) -> Iterable[RowDefn]:
    """Sort input data as specified by column definitions
    """
    sorting = [
        defn
        for defn in sorted(columns.values(), key=lambda defn: defn.sort_order)
        if defn.is_sortable
    ]

    if sorting:
        inputs = list(inputs)
        if any(row is None for row in inputs):
            raise TypeError(
                'Table row separators (row values of None) are incompatible with sorting. '
                'Either remove the row separators, or disable sorting.')

        inputs = sorted(inputs, key=lambda data: [
            defn.get_row_sort_key(data, columns)
            for defn in sorting
        ])

    return inputs


def render_table(columns: Dict[str, ColumnDefnSource],
                 inputs: InputsSource,
                 title: str = None,
                 **kwargs,
                 ) -> 'SingleTable':
    """Format a list of dicts as a table
    """
    columns = init_columns(columns)

    if callable(inputs):
        get_inputs = inputs
        inputs = get_inputs()

    inputs = sort_by_columns(columns, inputs)

    def format_row(data, columns):
        if data is not None:
            return [
                defn.get_row_formatted_value(data, columns)
                for defn in columns.values()
            ]

    rows = [format_row(data, columns) for data in inputs]

    header = [defn.get_title() for defn in columns.values()]
    rows.insert(0, header)

    table = SingleTable(rows, source_dicts=inputs, title=title)
    for attr, value in kwargs.items():
        setattr(table, attr, value)

    table.justify_columns = {
        index: defn.justify
        for index, defn in enumerate(columns.values())
    }

    return table


def display_table(columns: Dict[str, ColumnDefnSource],
                  inputs: InputsSource,
                  title: str = None,
                  file: IO = None,
                  **kwargs,
                  ) -> None:
    """Display a list of dicts as a table in the terminal
    """
    table = render_table(columns=columns, inputs=inputs, title=title, **kwargs)
    print(table.table, file=file)


def watch_table(columns: Dict[str, ColumnDefnSource],
                get_inputs: Callable[[], Iterable[RowDefn]],
                title: str = None,
                refresh: float = 5,
                on_screen_change: Callable[['WatchTable'], Any] = None,
                **kwargs,
                ) -> None:
    """Format a list of dicts as an interactive, periodically-refreshing table
    """
    table = WatchTable(
        columns=columns,
        get_inputs=get_inputs,
        title=title,
        refresh=refresh,
        on_screen_change=on_screen_change,
        **kwargs,
    )
    table.display_forever()


class WatchTable:
    def __init__(
        self,
        columns: Dict[str, ColumnDefnSource],
        get_inputs: Callable[[], Iterable[RowDefn]],
        title: str = None,
        refresh: float = 5,
        on_screen_change: Callable[['WatchTable'], Any] = None,
        swallow_errors: bool = True,
        has_border: bool = True,
        **kwargs,
    ):
        # Pre-init column definitions, so they needn't be initialized on every render
        self.columns = init_columns(columns)
        self.key = [col.name for col in self.columns.values() if col.is_key]

        self.get_inputs = get_inputs
        self.title = title
        self.refresh = refresh
        self.on_screen_change = on_screen_change
        self.swallow_errors = swallow_errors
        self.has_border = has_border
        self.table_kwargs = kwargs

        self.is_initialized: bool = False
        self.list_box: Optional[ColoredMultiColumnListBox] = None
        self.screen: Optional[Screen] = None
        self.frame: Optional[TableFrame] = None

        #: Errors raised while calling get_inputs, as (raised_at, traceback_str)
        self.encountered_errors: List[Tuple[datetime, str]] = []

        self._halt = False

    def stop(self):
        self._halt = True

    def display_forever(self):
        """Render the interactive table, and refresh forever
        """
        with original_stdstreams():
            while not self._halt:
                try:
                    self.initialize()

                    with ManagedScreen() as screen:
                        update_thread = IntervalThread(self.refresh, self._update, screen)
                        update_thread.start()

                        real_refresh = screen.refresh

                        def haltable_refresh(*args, **kwargs):
                            if self._halt:
                                raise StopApplication('Halt requested')
                            return real_refresh(*args, **kwargs)

                        screen.refresh = haltable_refresh

                        try:
                            self._watch(screen)
                        finally:
                            update_thread.stop()
                            self.screen = None
                except ResizeScreenError:
                    pass
                except KeyboardInterrupt:
                    self.stop()
                    if self.screen:
                        self.screen.close()
                    self.print_encountered_errors()
                    raise

    def initialize(self):
        if self.is_initialized:
            return

        self.list_box = ColoredMultiColumnListBox(
            name=self.title,
            key=self.key,
            height=Widget.FILL_FRAME,
            columns=[],
            options=[],
        )

        while True:
            attrs = self.get_widget_attrs()
            if attrs:
                break
            else:
                time.sleep(1)
                continue

        self.list_box.update_from_attrs(attrs)
        self.is_initialized = True

    def _update(self, screen: Screen):
        updated_attrs = self.get_widget_attrs()
        if updated_attrs is None:
            return

        prev_value = self.list_box.value
        self.list_box.update_from_attrs(updated_attrs)
        self.list_box.value = prev_value
        self.list_box.update(None)
        screen.force_update()

    def _watch(self, screen: Screen):
        self.screen = screen
        self.frame = self.create_frame()
        scene = Scene([self.frame], 0)

        if self.on_screen_change:
            self.on_screen_change(self)

        self.screen.play([scene], start_scene=scene, stop_on_resize=True)

    def create_frame(self):
        return TableFrame(self.screen, self.list_box, title=self.title, has_border=self.has_border)

    def print_encountered_errors(self):
        if self.encountered_errors:
            print_err = lambda *a, **k: print(*a, **k, file=sys.stderr)

            print_err(f'Encountered {len(self.encountered_errors)} error(s) while refreshing\n')

            for exc_at, tb in self.encountered_errors:
                print_err(f'At {exc_at}\n')
                print_err(tb, '\n\n')

    def get_widget_attrs(self) -> Dict[str, Any]:
        try:
            return self._get_widget_attrs()
        except (StopApplication, NextScene):
            raise
        except KeyboardInterrupt:
            self.stop()
        except Exception:
            if self.swallow_errors:
                tb = traceback.format_exc()
                exc_at = datetime.now()
                self.encountered_errors.append((exc_at, tb))
            else:
                self.stop()
                raise

    def _get_widget_attrs(self) -> Dict[str, Any]:
        inputs = self.get_inputs()
        table = render_table(
            columns=self.columns, inputs=inputs, title=self.title, **self.table_kwargs)
        return self.list_box.parse_attrs_from_table(table)


class AsyncTable:
    """An interactive, periodically-refreshing table powered by asyncio

    This table can be used as an async context manager — useful when you're
    doing things and want to show progress info about them; or this table can
    be used as an awaitable — useful when you're merely displaying status info
    about things going on outside of the program.

    Usage as context manager:

        async def my_provider():
            while True:
                yield await get_rows()
                await asyncio.sleep(1)

        async with AsyncTable(columns=my_columns, provider=my_provider()):
            while things_to_do:
                await do_a_thing()

    Usage as awaitable:

        async def my_provider():
            while True:
                yield await get_infos()
                await asyncio.sleep(1)

        await AsyncTable(columns=my_columns, provider=my_provider())

    """

    def __init__(
        self,
        columns: Dict[str, ColumnDefnSource],
        provider: Union[AsyncIterator[Iterable[RowDefn]], Callable[[], Awaitable[Iterable[RowDefn]]]],
        title: str = None,
        refresh: float = None,
        on_screen_change: Callable[['AsyncTable'], Any] = None,
        swallow_errors: bool = True,
        has_border: bool = True,
        tick: float = 0.05,
        **kwargs,
    ):
        # Pre-init column definitions, so they needn't be initialized on every render
        self.columns = init_columns(columns)
        self.key = [col.name for col in self.columns.values() if col.is_key]

        self.provider = provider
        self.title = title
        self.refresh = refresh
        self.on_screen_change = on_screen_change
        self.swallow_errors = swallow_errors
        self.has_border = has_border
        self.tick = tick
        self.table_kwargs = kwargs

        self.list_box = ColoredMultiColumnListBox(
            name=self.title,
            key=self.key,
            height=Widget.FILL_FRAME,
            columns=[],
            options=[],
        )

        #: Errors raised while calling get_inputs, as (raised_at, traceback_str)
        self.encountered_errors: List[Tuple[datetime, str]] = []

        self._task: Optional[asyncio.Task] = None

    def __await__(self):
        return self._display().__await__()

    async def __aenter__(self):
        self._task = asyncio.create_task(self._display())
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self._task:
            self._task.cancel()
            with suppress(asyncio.CancelledError):
                await self._task
            self._task = None

    async def _inputs_iterator(self):
        if inspect.iscoroutinefunction(self.provider):
            while True:
                yield await self.provider()

        elif inspect.isasyncgen(self.provider):
            async for inputs in self.provider:
                yield inputs

    async def _display(self):
        screen: Optional[Screen] = None
        update_task: Optional[asyncio.Task] = None
        are_inputs_depleted = asyncio.Event()
        got_first_input = asyncio.Event()
        did_initial_refresh_screen_callback = False

        async def periodic_update():
            async for inputs in self._inputs_iterator():
                table = render_table(
                    columns=self.columns,
                    inputs=inputs,
                    title=self.title,
                    **self.table_kwargs,
                )
                self.list_box.update_from_table(table)

                self.list_box.update(None)
                if screen is not None:
                    screen.force_update()

                got_first_input.set()

                if self.refresh is not None:
                    await asyncio.sleep(self.refresh)

            # When inputs are depleted, end display
            are_inputs_depleted.set()

        try:
            with original_stdstreams():
                while not are_inputs_depleted.is_set():
                    with ManagedScreen() as screen:
                        frame = TableFrame(screen, self.list_box,
                                           title=self.title,
                                           has_border=self.has_border)
                        scene = Scene([frame], -1)
                        screen.set_scenes([scene], start_scene=scene)

                        if update_task is None:
                            update_task = asyncio.create_task(periodic_update())

                        if self.on_screen_change:
                            self.on_screen_change(self)

                        while not screen.has_resized() and not are_inputs_depleted.is_set():
                            screen.draw_next_frame()

                            if got_first_input.is_set() and not did_initial_refresh_screen_callback:
                                if self.on_screen_change:
                                    self.on_screen_change(self)
                                did_initial_refresh_screen_callback = True

                            await asyncio.sleep(self.tick)
        finally:
            if screen is not None:
                screen.close()

            if update_task:
                update_task.cancel()
                with suppress(asyncio.CancelledError):
                    await update_task


class IntervalThread(Thread):
    """Invokes a callback every N seconds, until stop() is called"""

    def __init__(self, interval: float, callback: Callable, *args, **kwargs):
        Thread.__init__(self)
        self.interval = interval
        self.callback = callback
        self.args = args
        self.kwargs = kwargs
        self._cancel = Event()

    def run(self):
        while not self._cancel.wait(self.interval):
            self.callback(*self.args, **self.kwargs)

    def stop(self):
        self._cancel.set()


class TableFrame(Frame):
    def __init__(self, screen: Screen, list_box: ColoredMultiColumnListBox, **kwargs):
        super().__init__(screen, screen.height, screen.width, **kwargs)

        layout = Layout([100], fill_frame=True)
        self.add_layout(layout)

        layout.add_widget(list_box)
        self.set_theme('monochrome')

        self.fix()


class TableSeparatorMixin:
    """Allow a row value of None to print a horizontal border

    By default, AsciiTables only allow a single footer row. This mixin also allows
    the inner_footing_row_border instance var to be set to an integer describing
    the number of footer rows there are (i.e. the number of rows to print after
    the footer border)

    XXX: this mixin now does more things — rename and redocument
    """

    show_header_row: bool

    table_data: List[Dict[str, Any]]

    inner_column_border: bool
    inner_footing_row_border: bool
    inner_heading_row_border: bool
    inner_row_border: bool
    outer_border: bool

    horizontal_border: Callable[[str, Collection[int]], Iterable[str]]
    gen_row_lines: Callable[[Dict, str, Collection[int], int], Iterable[str]]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.show_header_row = True

    def gen_table(self, inner_widths, inner_heights, outer_widths):
        """Combine everything and yield every line of the entire table with borders.

        :param iter inner_widths: List of widths (no padding) for each column.
        :param iter inner_heights: List of heights (no padding) for each row.
        :param iter outer_widths: List of widths (with padding) for each column.
        :return:
        """
        # Yield top border.
        if self.outer_border:
            yield self.horizontal_border('top', outer_widths)

        # Yield table body.
        row_count = len(self.table_data)
        last_row_index = row_count - 1
        before_footer_row_index = row_count - (self.inner_footing_row_border + 1)

        for i, row in enumerate(self.table_data):
            # Yield the row line by line (e.g. multi-line rows).
            if self.inner_heading_row_border and i == 0:
                if not self.show_header_row:
                    continue
                style = 'heading'
            elif self.inner_footing_row_border and i == last_row_index:
                style = 'footing'
            else:
                style = 'row'

            if row is None:
                yield self.horizontal_border('row', outer_widths)
                continue

            for line in self.gen_row_lines(row, style, inner_widths, inner_heights[i]):
                yield line

            # If this is the last row then break. No separator needed.
            if i == last_row_index:
                break

            # Yield heading separator.
            if self.inner_heading_row_border and i == 0:
                yield self.horizontal_border('heading', outer_widths)

            # Yield footing separator.
            elif self.inner_footing_row_border and i == before_footer_row_index:
                yield self.horizontal_border('footing', outer_widths)

            # Yield row separator.
            elif self.inner_row_border:
                yield self.horizontal_border('row', outer_widths)

        # Yield bottom border.
        if self.outer_border:
            yield self.horizontal_border('bottom', outer_widths)

    @property
    def table(self):
        """Return a large string of the entire table ready to be printed to the terminal."""
        dimensions = max_dimensions_ignore_none(self.table_data, self.padding_left, self.padding_right)[:3]
        return flatten(self.gen_table(*dimensions))


def max_dimensions_ignore_none(table_data, padding_left=0, padding_right=0, padding_top=0, padding_bottom=0):
    """Get maximum widths of each column and maximum height of each row.

    NOTE: this method skips None values, for use with TableSeparatorMixin

    :param iter table_data: List of list of strings (unmodified table data).
    :param int padding_left: Number of space chars on left side of cell.
    :param int padding_right: Number of space chars on right side of cell.
    :param int padding_top: Number of empty lines on top side of cell.
    :param int padding_bottom: Number of empty lines on bottom side of cell.

    :return: 4-item tuple of n-item lists. Inner column widths and row heights, outer column widths and row heights.
    :rtype: tuple
    """
    inner_widths = [0] * (max(len(r) if r is not None else 0 for r in table_data) if table_data else 0)
    inner_heights = [0] * len(table_data)

    # Find max width and heights.
    for j, row in enumerate(table_data):
        if row is None:
            continue

        for i, cell in enumerate(row):
            if not hasattr(cell, 'count') or not hasattr(cell, 'splitlines'):
                cell = str(cell)
            if not cell:
                continue
            inner_heights[j] = max(inner_heights[j], cell.count('\n') + 1)
            inner_widths[i] = max(inner_widths[i], *[visible_width(l) for l in cell.splitlines()])

    # Calculate with padding.
    outer_widths = [padding_left + i + padding_right for i in inner_widths]
    outer_heights = [padding_top + i + padding_bottom for i in inner_heights]

    return inner_widths, inner_heights, outer_widths, outer_heights


class SingleTable(TableSeparatorMixin, terminaltables.SingleTable):
    def __init__(self, table_data, source_dicts=None, title=None):
        """Constructor.

        :param iter table_data: List (empty or list of lists of strings) representing the table.
        :param title: Optional title to show within the top border of the table.
        """
        self.source_dicts = source_dicts
        super().__init__(table_data, title=title)


# Monkey-patch our fixed visible_width
import terminaltables.width_and_alignment

terminaltables.build.visible_width = visible_width
terminaltables.width_and_alignment.visible_width = visible_width
